// Lzsa1Decode:
#ifndef FLAT32
//   input   ds:si=inStream, es:di=outStream
//   output  outStream[], ds:si, es:di
	.code16
#define AX	%ax
#define BX	%bx
#define SI	%si
#define DI	%di
#else
//   input   esi=inStream, edi=outStream
//   output  outStream[], ds:esi, es:edi
	.code32
#define AX	%eax
#define BX	%ebx
#define SI	%esi
#define DI	%edi
#endif

MATCH_RUN_LEN		=	15
LITERALS_RUN_LEN	=	7
MIN_MATCH_SIZE		=	3
MIN_LITERALS_SIZE	=	0

#define PACKED_ONLY			// assume no copy block, optional
//#define PARANOIA			// cover rare cases, optional
 
lzsa1main:
#ifdef PARANOIA
	cld
#endif
#ifdef FLAT16OUT
#define RAW_FORMAT
#endif
#ifndef RAW_FORMAT
# if defined(PARANOIA) && !defined(FLAT32) && !defined(FLAT16)
	xorw	%cx, %cx
	call	normalize
# endif
# ifndef NO_LZSA1_HEADER
	lodsw
	cmpw	$0x9E7B, %ax	// magic
	jne	lzsa1main
	lodsb
	cmpb	$0, %al		// lzsa1
	jne	lzsa1main
# endif
lzsa1block:			// uncompress chunk
	lodsw			// block size
	xchgw	%ax, %cx
	lodsb
# ifndef PACKED_ONLY
	orb	%al, %al
	jns	lzsa1compressed
#  if !defined(FLAT32) && !defined(FLAT16OUT)
	movw	%cx, %dx
	movb	$0, %cl
	movb	$0, %dh
copytail:
	call	lzsa1movStr
	xchg	%dx, %cx
	incw	%cx
	loop	copytail
#  else
	movsb			// copy block
copylp:
	movsb			// copy block
	loop	copylp		// handle 64K case
#  endif
	jmp	lzsa1block	// end of block
lzsa1compressed:
	jne	lzsa1chunk	// 64Kb block
# endif
	jcxz	lzsa1quit	// bail if we hit EOD
	pushw	%cx
# if !defined(FLAT32) && !defined(FLAT16)
	xorw	%cx, %cx
	call	normalize
# define NeedNormalize
# endif
	popw	%dx
	addw	%si, %dx
#endif
lzsa1chunk:			// uncompress chunk
	lodsb			// get token O|LLL|MMMM
	movb	%al, %bl	// keep token in bl
	shrb	$4, %al		// shift literals length into place
	movw	$LITERALS_RUN_LEN*256+MIN_LITERALS_SIZE, %cx
	call	lzsa1len	// %ch = LITERALS_RUN_LEN
#if defined(NeedNormalize)
	call	lzsa1movLit	// copy %cx literals from %ds:%si to %es:%di
#else
	rep	movsb		// copy %cx literals from %ds:%si to %es:%di
#endif
#ifndef RAW_FORMAT
	cmpw	%dx, %si
	jae	lzsa1block	// bail if we hit EOD
#endif
#ifdef FLAT32
	orl	$-1, %eax
#else
	movb	$-1, %ah
#endif
	testb	%bl, %bl	// check match offset size in token (O bit)
	jns	lzsa1ShortOfs
	lodsw
	.byte	0x3C		// mask lodsb with cmpb $0xAC, %al
lzsa1ShortOfs:
	lodsb
	xchg	AX, BX		// %bx: match offset  %ax: original token
	movw	$MATCH_RUN_LEN*256+MIN_MATCH_SIZE, %cx
	call	lzsa1len
#ifdef RAW_FORMAT
	jcxz	lzsa1quit	// bail if we hit EOD
#endif
#if !defined(FLAT32) && !defined(FLAT16OUT)
	pushw	%ds
	pushw	%si
	movw	%di, %si
	addw	%bx, %si
	movw	%es, %ax
	jc	axok
	subb	$0x10, %ah
axok:
.macro norm	reg
	movw	%si, \reg
	subw	%si, %dx
	andw	$0xF, %si
	addw	%si, %dx
	shrw	$4, \reg
	addw	\reg, %ax
	movw	%ax, %ds
	movw	%di, \reg
	andw	$0xF, %di
	shrw	$4, \reg
	movw	%es, %ax
	addw	\reg, %ax
	movw	%ax, %es
.endm 
	pushw	%dx
# if defined(NeedNormalize) || defined(PARANOIA)
	call	lzsa1movStr	// copy string
# else
	norm	%bp
	rep movsb
# endif
	popw	%dx
	popw	%si
	popw	%ds
#else
	xchg	AX, SI		// save %si	
	lea	(BX,DI), SI
	rep movsb	%es:(SI), %es:(DI)
	xchg	AX, SI		// restore %si	
#endif
	jmp	lzsa1chunk

lzsa1len:			// get length in %ecx
	andb	%ch, %al
	cbw			// clear %ah
	cmpb	%ch, %al
	jne	lzsa1minNumber	// S=0-6, L=0-14
	lodsb
	addb	%ch, %cl
lzsa1minNumber:
	addb	%cl, %al
	jnc	lzsa1gotNumber  // 0-255
	movb	%al, %ah	// S=256-1791, L=256-3839 or S=256-511, L=256-511
	jne	lzsa1midNumber
	lodsw			// 0-65535
	.byte	0x3C		// mask lodsb with cmpb $0xAC, %al
lzsa1midNumber:
	lodsb
lzsa1gotNumber:
	xchgw	%ax, %cx
lzsa1quit:
	ret

#if defined(NeedNormalize) || defined(PARANOIA)
# if defined(PARANOIA)
lzsa1movlp:
	decw	%ch
	rep movsb
	incw	%ch
# endif
normalize:
lzsa1movLit:
	movw	%ds, %ax
lzsa1movStr:
	norm	%bp
# if defined(PARANOIA)
	cmpb	$0xFF, %ch	// catch FFFX case
	je	lzsa1movlp
# endif
	rep movsb
	ret
#endif
